import subprocess
import time
import calendar
import os
import random
import sqlite3
import xml.etree.ElementTree as ET
import signal
import sys
import argparse
import threading
import fcntl
import errno

class StreamManager:
    def __init__(self, stream_name, shuffle=False):
        # Instance-specific variables
        self.stream_name = stream_name
        self.shuffle = shuffle
        self.epg_buffer_hours = 72
        self.time_offset = 0
        self.epg_sync_interval = 7200
        self.duration_cache_db = "/mnt/2tb/247/duration_cache.db"
        self.hls_dir = "/home/http/www/live"
        self.epg_file = "/home/http/www/247.xml"
        self.epg_lock_file = "/tmp/247_epg.lock"
        self.ffmpeg_pid = None
        self.last_epg_sync_check = 0
        self.initial_epg_generated = False
        self.currently_playing_video = None
        self.epg_generation_started = False
        self.playlist = f"/mnt/2tb/247/{self.stream_name}.txt"
        self.temp_playlist = f"/tmp/{self.stream_name}_current.txt"
        self.shuffle_playlist = f"/tmp/{self.stream_name}_shuffled.txt"
        self.temp_epg_file = f"/tmp/{self.stream_name}_epg.xml"
        self.timestamped_playlist = f"/tmp/{self.stream_name}_timestamped.txt"
        self.channel_id = f"{self.stream_name}247"

        self.init_db()

        # Validate playlist
        if not os.path.isfile(self.playlist):
            print(f"Error: Playlist not found at {self.playlist}")
            sys.exit(1)
        if os.stat(self.playlist).st_size == 0:
            print(f"Error: Playlist is empty at {self.playlist}")
            sys.exit(1)

        # Prepare playlist (this needs to happen before finding the resume point)
        self.prepare_playlist()

        # Add a small delay to allow the file system to catch up
        time.sleep(0.1)  # Adjust delay as needed

        # Retrieve last played video information
        conn = sqlite3.connect(self.duration_cache_db)
        cursor = conn.cursor()
        cursor.execute(
            "SELECT video_path, timestamp FROM last_played WHERE stream_name = ?",
            (self.stream_name,),
        )
        result = cursor.fetchone()
        conn.close()

        if result:
            last_played_video, last_played_timestamp = result
            self.currently_playing_video = last_played_video
            print(
                f"[{self.stream_name}] Resuming from: {self.currently_playing_video} (timestamp: {last_played_timestamp})"
            )
        else:
            print(
                f"[{self.stream_name}] No last played entry found. Starting from the beginning."
            )

        # Set initial timestamp, if resuming, grab from playlist
        if self.currently_playing_video:
            with open(self.playback_playlist, "r") as f:
                for line in f:
                    video_path, ts = line.split(",", 1)
                    video_path = video_path.replace("{{COMMA}}", ",")
                    if video_path == self.currently_playing_video:
                        current_timestamp = int(ts)
                        break
        else:
            current_timestamp = int(time.time())

        # Generate EPG *after* finding the resume point
        if not self.initial_epg_generated:
            self.generate_epg()
            self.initial_epg_generated = True

    def init_db(self):
        conn = sqlite3.connect(self.duration_cache_db)
        cursor = conn.cursor()
        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS video_durations (
                filepath TEXT PRIMARY KEY,
                duration REAL,
                last_updated INTEGER
            )
            """
        )
        # Create a table for last played information
        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS last_played (
                stream_name TEXT PRIMARY KEY,
                video_path TEXT,
                timestamp INTEGER
            )
            """
        )
        conn.commit()
        conn.close()

    def cleanup(self):
        print(f"[{self.stream_name}] Cleaning up...")
        if self.ffmpeg_pid:
            print(
                f"[{self.stream_name}] Terminating FFmpeg process (PID: {self.ffmpeg_pid})..."
            )
            try:
                # Check if the process exists before sending signals
                os.kill(self.ffmpeg_pid, 0)

                # Terminate the process group
                os.kill(-self.ffmpeg_pid, signal.SIGTERM)

                # Wait for the process to terminate (optional, with timeout)
                self.ffmpeg_process.wait(timeout=5)

            except OSError as e:
                if e.errno != errno.ESRCH:  # No such process
                    print(
                        f"[{self.stream_name}] Error during FFmpeg process termination: {e}"
                    )
            except subprocess.TimeoutExpired:
                print(
                    f"[{self.stream_name}] FFmpeg process did not terminate in time. Killing forcefully..."
                )
                os.kill(-self.ffmpeg_pid, signal.SIGKILL)
        for f in [
            self.temp_playlist,
            self.shuffle_playlist,
            self.temp_epg_file,
            self.timestamped_playlist,
        ]:
            try:
                os.remove(f)
                print(f"[{self.stream_name}] Removed temporary file: {f}")
            except OSError as e:
                print(f"[{self.stream_name}] Error removing temporary file {f}: {e}")

    def extract_title(self, filepath):
        filename = os.path.basename(filepath)
        filename = os.path.splitext(filename)[0]

        import re

        # Remove brackets and their contents
        filename = re.sub(r"\[.*?\]", "", filename)
        # Remove parentheses and their contents
        filename = re.sub(r"\(.*?\)", "", filename)
        # Remove multiple spaces
        filename = re.sub(r"\s+", " ", filename)
        # Trim whitespace
        filename = filename.strip()

        # Escape &, <, >, ", and ' for XML
        filename = filename.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;").replace("\'", "&apos;").replace("\"", "&quot;")

        return filename


    def generate_timestamp_utc(self, time_string):
        if not time_string or not isinstance(time_string, str):
            return None
        try:
            # Remove the timezone offset if it exists
            time_string = time_string.split(" ")[0]
            # Parse the time string assuming it's in UTC
            utc_time_struct = time.strptime(time_string, "%Y%m%d%H%M%S")
            # Convert to UTC timestamp
            utc_timestamp = calendar.timegm(utc_time_struct)
            return int(utc_timestamp)
        except ValueError as e:
            print(f"Error: Could not parse time string {time_string}: {e}")
            return None

    def get_video_duration(self, video_path):
        conn = sqlite3.connect(self.duration_cache_db)
        cursor = conn.cursor()

        # Check if the duration is in the cache and not older than 30 days
        cursor.execute(
            "SELECT duration FROM video_durations WHERE filepath = ? AND last_updated > ?",
            (video_path, int(time.time() - 30 * 24 * 3600)),
        )
        result = cursor.fetchone()

        if result:
            conn.close()
            return result[0]

        try:
            duration = (
                subprocess.check_output(
                    [
                        "ffprobe",
                        "-v",
                        "error",
                        "-select_streams",
                        "v:0",
                        "-show_entries",
                        "format=duration",
                        "-of",
                        "default=noprint_wrappers=1:nokey=1",
                        video_path,
                    ],
                    stderr=subprocess.DEVNULL,
                )
                .decode()
                .strip()
            )
        except subprocess.CalledProcessError:
            duration = "1"  # Default to 20 minutes

        if not duration or not duration.replace(".", "", 1).isdigit():
            duration = "1200"

        # Store the duration in the database
        cursor.execute(
            "INSERT OR REPLACE INTO video_durations (filepath, duration, last_updated) VALUES (?, ?, ?)",
            (video_path, float(duration), int(time.time())),
        )
        conn.commit()
        conn.close()

        return float(duration)

    def get_total_epg_duration(self):
        current_time_utc = int(time.time())

        if os.path.isfile(self.epg_file):
            try:
                tree = ET.parse(self.epg_file)
                root = tree.getroot()
                latest_stop_utc = None

                for programme in root.findall(
                    f".//programme[@channel='{self.channel_id}']"
                ):
                    stop_time_str = programme.get("stop")
                    if stop_time_str:
                        stop_time_str = stop_time_str.split(" ")[0]  # Remove timezone
                        stop_timestamp_utc = self.generate_timestamp_utc(stop_time_str)

                        if not latest_stop_utc or stop_timestamp_utc > latest_stop_utc:
                            latest_stop_utc = stop_timestamp_utc

                if latest_stop_utc:
                    return latest_stop_utc - current_time_utc

            except ET.ParseError:
                print(f"[{self.stream_name}] Error parsing EPG XML file.")

        return 0

    def generate_epg(self):
        print(f"[{self.stream_name}] Generating EPG...")

        # Work exclusively in UTC
        current_timestamp_utc = int(time.time())

        target_duration = self.epg_buffer_hours * 3600
        total_duration = 0
        epg_data = []

        playlist_entries = []
        with open(self.playback_playlist, "r") as f:
            for line in f:
                if line.strip():
                    video_path, _ = line.strip().split(",", 1)
                    video_path = video_path.replace("{{COMMA}}", ",")
                    playlist_entries.append(video_path)

        playlist_index = 0
        if self.currently_playing_video:
            # Find the index of the currently playing video
            try:
                playlist_index = playlist_entries.index(self.currently_playing_video)
            except ValueError:
                print(
                    f"[{self.stream_name}] Currently playing video not found in playlist."
                )
                playlist_index = 0

        while total_duration < target_duration:
            if playlist_index >= len(playlist_entries):
                playlist_index = 0  # Reset index to loop from the beginning

            video_path = playlist_entries[playlist_index]

            if not video_path:
                playlist_index += 1
                continue

            duration = int(self.get_video_duration(video_path))

            # Use current_timestamp_utc for start and stop times (UTC)
            start_time_utc = time.strftime(
                "%Y%m%d%H%M%S", time.gmtime(current_timestamp_utc)
            )
            stop_time_utc = time.strftime(
                "%Y%m%d%H%M%S", time.gmtime(current_timestamp_utc + duration)
            )
            title = self.extract_title(video_path)

            epg_data.append(
                f"""    <programme start="{start_time_utc} +0000" stop="{stop_time_utc} +0000" channel="{self.channel_id}">
            <title>{title}</title>
        </programme>"""
            )

            current_timestamp_utc += duration  # Advance by duration
            total_duration += duration
            playlist_index += 1

        # Create the root element for the temporary EPG
        temp_epg_root = ET.Element("tv")

        # Create the channel element and add it to the root
        channel_element = ET.SubElement(temp_epg_root, "channel", id=self.channel_id)
        ET.SubElement(channel_element, "display-name").text = self.channel_id

        # Add programme elements to the root
        for item in epg_data:
            # Parse the string into an ElementTree element
            programme_element = ET.fromstring(item)
            temp_epg_root.append(programme_element)

        # Write the temporary EPG to file
        temp_tree = ET.ElementTree(temp_epg_root)
        temp_tree.write(self.temp_epg_file, encoding="utf-8", xml_declaration=True)

        # Acquire a lock on the EPG lock file
        epg_lock = open(self.epg_lock_file, "a+")  # Use r+ mode for reading and writing
        try:
            fcntl.flock(epg_lock, fcntl.LOCK_EX)  # Acquire an exclusive lock

            # --- START OF CRITICAL SECTION (Only one instance can be here at a time) ---

            # Parse the EPG file or create a new one if it doesn't exist
            if os.path.isfile(self.epg_file):
                try:
                    tree = ET.parse(self.epg_file)
                    root = tree.getroot()
                except ET.ParseError:
                    print(
                        f"[{self.stream_name}] EPG file was corrupted. Creating a new one."
                    )
                    root = ET.Element("tv")
            else:
                print(f"[{self.stream_name}] EPG file not found. Creating a new one.")
                root = ET.Element("tv")

            # Find or create the channel element for this stream
            channel_element = root.find(f"./channel[@id='{self.channel_id}']")
            if channel_element is None:
                channel_element = ET.SubElement(root, "channel", id=self.channel_id)
                ET.SubElement(channel_element, "display-name").text = self.channel_id

            # Remove existing programmes for this channel
            for programme in root.findall(
                f".//programme[@channel='{self.channel_id}']"
            ):
                root.remove(programme)

            # Parse the temporary EPG file
            temp_tree = ET.parse(self.temp_epg_file)
            temp_root = temp_tree.getroot()

            # Add new programme entries for this channel from the temporary EPG
            for programme in temp_root.findall(
                f".//programme[@channel='{self.channel_id}']"
            ):
                root.append(programme)

            # Overwrite the EPG file with the updated content
            tree = ET.ElementTree(root)
            tree.write(self.epg_file, encoding="utf-8", xml_declaration=True)

            # --- END OF CRITICAL SECTION ---

        finally:
            fcntl.flock(epg_lock, fcntl.LOCK_UN)  # Release the lock
            epg_lock.close()

        self.initial_epg_generated = True
        self.epg_generation_started = True


    def process_video(self, video_path, expected_start_timestamp):
        if not os.path.isfile(video_path):
            print(f"[{self.stream_name}] Error: File not found - {video_path}")
            return 1
        print(f"[{self.stream_name}] Streaming: {video_path}")
        duration = int(self.get_video_duration(video_path))

        # Connect to the database and update last_played table
        conn = sqlite3.connect(self.duration_cache_db)
        cursor = conn.cursor()
        cursor.execute(
            "INSERT OR REPLACE INTO last_played (stream_name, video_path, timestamp) VALUES (?, ?, ?)",
            (self.stream_name, video_path, int(time.time())),
        )
        conn.commit()
        conn.close()

        ffmpeg_cmd = [
            "ffmpeg",
            "-nostdin",
            "-loglevel",
            "error",
            "-re",
            "-y",
            "-i",
            video_path,
            "-c",
            "copy",
            "-f",
            "hls",
            "-hls_time",
            "10",
            "-sn",
            "-hls_list_size",
            "5",
            "-hls_flags",
            "delete_segments+append_list",
            "-hls_segment_filename",
            f"{self.hls_dir}/{self.stream_name}_%03d.ts",
            f"{self.hls_dir}/{self.stream_name}.m3u8",
        ]
        try:
            self.ffmpeg_process = subprocess.Popen(
                ffmpeg_cmd, stderr=subprocess.PIPE, preexec_fn=os.setsid
            )
            self.ffmpeg_pid = self.ffmpeg_process.pid
            print(
                f"[{self.stream_name}] FFmpeg process started (PID: {self.ffmpeg_pid})"
            )
            while self.ffmpeg_process.poll() is None:
                time.sleep(1)
            stdout, stderr = self.ffmpeg_process.communicate()
            ffmpeg_exit_status = self.ffmpeg_process.returncode
            if ffmpeg_exit_status != 0:
                print(
                    f"[{self.stream_name}] FFmpeg process completed with an error (exit code: {ffmpeg_exit_status})"
                )
                print(f"[{self.stream_name}] FFmpeg stderr: {stderr.decode()}")
                return 1
            else:
                print(f"[{self.stream_name}] FFmpeg process completed successfully.")
            return 0
        except FileNotFoundError:
            print(
                f"[{self.stream_name}] FFmpeg not found. Ensure it is installed and in your PATH."
            )
            return 1
        except Exception as e:
            print(
                f"[{self.stream_name}] Error during FFmpeg execution for video: {video_path}"
            )
            print(f"[{self.stream_name}] Exception: {e}")
            return 1

    def check_epg_regeneration(self):
        if not self.initial_epg_generated:
            return

        # Work exclusively in UTC
        current_time_utc = int(time.time())

        # Check if enough time has passed since the last sync check
        if current_time_utc - self.last_epg_sync_check >= self.epg_sync_interval:
            print(
                f"[{self.stream_name}] {self.epg_sync_interval} seconds have passed. Regenerating EPG..."
            )
            self.generate_epg()
            self.last_epg_sync_check = current_time_utc
        else:
            # Check if EPG buffer is low
            current_epg_duration = self.get_total_epg_duration()
            if current_epg_duration < self.epg_buffer_hours * 3600:
                print(f"[{self.stream_name}] EPG buffer low. Regenerating EPG...")
                self.generate_epg()
                self.last_epg_sync_check = current_time_utc

    def prepare_playlist(self):
        # Create a temporary playlist without shuffling initially
        with open(self.playlist, "r") as f_in, open(self.temp_playlist, "w") as f_out:
            for line in f_in:
                if line.strip():
                    f_out.write(line)

        if os.stat(self.temp_playlist).st_size == 0:
            print(f"[{self.stream_name}] Error: No valid entries found in playlist.")
            with open(self.playlist, "r") as f:
                print(f.read())
            sys.exit(1)

        # Shuffle for playback if enabled
        if self.shuffle:
            print(f"[{self.stream_name}] Shuffling playlist...")
            with open(self.temp_playlist, "r") as f_in, open(
                self.shuffle_playlist, "w"
            ) as f_out:
                lines = f_in.readlines()
                random.shuffle(lines)
                f_out.writelines(lines)
            self.playback_playlist = self.shuffle_playlist
        else:
            self.playback_playlist = self.temp_playlist

        # Attempt to retrieve cached timestamps else cache default timestamps:
        playlist_with_timestamps = []

        try:
            with open(self.timestamped_playlist, "r") as f:
                for entry in f:
                    if entry.strip():
                        video_path, cached_timestamp = entry.split(",", 1)
                        playlist_with_timestamps.append(
                            (video_path, int(cached_timestamp))
                        )
                print(
                    f"[{self.stream_name}] Loaded {len(playlist_with_timestamps)} entries from cache."
                )
        except FileNotFoundError:
            print(f"[{self.stream_name}] No cached timestamp file found.")
            # Calculate and store expected start timestamps
        current_timestamp_utc = (
            int(time.time()) + self.time_offset
        )  # Now the offset is added once at the start

        with open(self.playback_playlist, "r") as f:
            for video_path in f:
                video_path = video_path.strip()
                if video_path:
                    duration = int(self.get_video_duration(video_path))
                    playlist_with_timestamps.append(
                        (video_path, current_timestamp_utc)
                    )
                    current_timestamp_utc += duration  # Increment in UTC

        # Write the timestamped playlist to a new temporary file
        with open(self.timestamped_playlist, "w") as f:
            for video_path, timestamp in playlist_with_timestamps:
                # Replace commas with a unique delimiter that won't appear in filenames
                escaped_path = video_path.replace(",", "{{COMMA}}")
                f.write(f"{escaped_path},{timestamp}\n")

        self.playback_playlist = self.timestamped_playlist

    def run(self):
        print(f"[{self.stream_name}] Starting stream")
        print(f"[{self.stream_name}] Playlist: {self.playlist}")
        print(f"[{self.stream_name}] HLS output: {self.hls_dir}")
        if self.shuffle:
            print(f"[{self.stream_name}] Shuffle mode enabled.")
        with open(self.playlist, "r") as f:
            print(
                f"[{self.stream_name}] Original playlist entries: {len([line for line in f if line.strip()])}"
            )
        while True:
            playlist_entries = []

            # Use self.playback_playlist (either temp_playlist or shuffle_playlist)
            self.prepare_playlist()
            with open(self.playback_playlist, "r") as f:
                for line in f:
                    playlist_entries.append(line.strip())

            # Initialize playlist_index to 0 (default: start from the beginning)
            playlist_index = 0

            # If it has a start video, find it in the playlist.
            if self.currently_playing_video:
                for ix, line in enumerate(playlist_entries):
                    video_path, ts = line.split(",", 1)
                    video_path = video_path.replace("{{COMMA}}", ",")
                    if video_path == self.currently_playing_video:
                        playlist_index = ix
                        break
            print(f"[{self.stream_name}] Beginning inner loop.")
            while True:
                video_path, expected_start_timestamp = playlist_entries[
                    playlist_index % len(playlist_entries)
                ].split(",", 1)
                video_path = video_path.replace("{{COMMA}}", ",")
                self.currently_playing_video = video_path
                expected_start_timestamp = int(expected_start_timestamp)

                # Perform check_epg_regeneration() before starting each video
                self.check_epg_regeneration()
                return_code = self.process_video(video_path, expected_start_timestamp)

                if return_code != 0:
                    playlist_index += 1
                    continue

                playlist_index += 1
                if playlist_index >= len(playlist_entries):
                    break

            # Check if EPG needs regeneration at the end of each loop iteration
            self.check_epg_regeneration()
            print(
                f"[{self.stream_name}] Finished EPG check and buffer, moving to next loop."
            )

def main():
    parser = argparse.ArgumentParser(description="Stream Manager with EPG")
    parser.add_argument(
        "-s", "--shuffle", action="store_true", help="Enable shuffle mode"
    )
    parser.add_argument("stream_name", help="Name of the stream")
    args = parser.parse_args()

    manager = StreamManager(args.stream_name, args.shuffle)

    def signal_handler(signum, frame):
        manager.cleanup()
        sys.exit(0)

    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    manager.run()

if __name__ == "__main__":
    main()
